#include "common.h"

static char UNUSED rcsid[] = "$Id: lcc-libomp-1.0.c 9999 2020-12-31 00:00:00Z harald $";

#ifdef HAVE_DLFCN_H
# define __USE_GNU
# include <dlfcn.h>
# undef  __USE_GNU
#endif
#ifdef HAVE_STDIO_H
# include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif
#if !defined(HAVE__SYNC_FETCH_AND_ADD)
# ifdef HAVE_PTHREAD_H
#  include <pthread.h>
# endif
#endif

#include "wrapper.h"
#include "omp-common.h"

#define DEBUG

#if defined(PIC)

struct openmp_task_st
{
	void *p1;
	void *p2;
	long long task_ctr; /* assigned task counter */
};

/*
The nowait issue:

Some OpenMP clauses may include an optional directive called "nowait".
When specified, a thread won't wait for the others when finishing its
parallel region. Our callback pointers (pardo_uf, do_uf, par_uf,
par_single, par_sections) are set by each thread in a parallel region.
While all threads are simultaneously in the same region, there's no problem.
However, if just one thread moves to the following parallel region and
modifies this variable, the other threads end up invoking a wrong callback.
So far, I've found this problem in DO clauses, but "nowait" can be specified
in many others.

Which of the callback pointers below have to be changed for an array of
function pointers?  All of them, maybe?
*/

/* Pointer to the user function called by a PARALLEL REGION */
/* FIXME: Array of function pointers indexed by thread? (nowait issue) */
static void (*par_uf)(void*) = NULL;

/*
	callme_par (void *)
	With the same header as the routine to be called by the SMP runtime, just
	acts as a trampoline to this call. Each thread runs the very same routine
	with different params.
*/
static void callme_par (void *p1)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d callme_par: par_uf()=%p p1=%p\n", THREADID, par_uf, p1);
#endif

	if (par_uf == NULL)
	{
		fprintf (stderr, PACKAGE_NAME": Error! Invalid initialization of 'par_uf'\n");
		exit (0);
	}

	Extrae_OpenMP_UF_Entry (par_uf);
	Backend_setInInstrumentation (THREADID, FALSE); /* We're about to execute user code */
	par_uf (p1);
	Backend_setInInstrumentation (THREADID, TRUE); /* We're back to execute OpenMP code */
	Extrae_OpenMP_UF_Exit ();
}

//static void (*GOMP_parallel_real)(void*,void*,unsigned,unsigned) = NULL;
// ???

//static void (*GOMP_parallel_start_real)(void*,void*,unsigned) = NULL;
static void (*__omp_parallel_start_real)(void*, void*) = NULL;

//static void (*GOMP_parallel_end_real)(void) = NULL;
static void (*__omp_parallel_end_real)(void) = NULL;

//static void (*GOMP_barrier_real)(void) = NULL;
static void (*__omp_barrier_real)(void) = NULL;

//static void (*GOMP_critical_name_start_real)(void**) = NULL;
static void (*__omp_critical_name_enter_real)(void*) = NULL;

//static void (*GOMP_critical_name_end_real)(void**) = NULL;
static void (*__omp_critical_name_exit_real)(void*) = NULL;

//static void (*GOMP_critical_start_real)(void) = NULL;
static void (*__omp_critical_enter_real)(void) = NULL;

//static void (*GOMP_critical_end_real)(void) = NULL;
static void (*__omp_critical_exit_real)(void) = NULL;

//static void (*GOMP_atomic_start_real)(void) = NULL;
static void (*__omp_atomic_enter_real)(void) = NULL;

//static void (*GOMP_atomic_end_real)(void) = NULL;
static void (*__omp_atomic_exit_real)(void) = NULL;

//static void (*GOMP_parallel_loop_static_start_real)(void*,void*,unsigned, long, long, long, long) = NULL;
// ???

//static void (*GOMP_parallel_loop_runtime_start_real)(void*,void*,unsigned, long, long, long, long) = NULL;
// ???

//static void (*GOMP_parallel_loop_dynamic_start_real)(void*,void*,unsigned, long, long, long, long) = NULL;
// ???

//static void (*GOMP_parallel_loop_guided_start_real)(void*,void*,unsigned, long, long, long, long) = NULL;
// ???

//static int (*GOMP_loop_static_next_real)(long*,long*) = NULL;
static int (*__omp_loop_static_next_real)(int*,int*) = NULL;

//static int (*GOMP_loop_runtime_next_real)(long*,long*) = NULL;
static void (*__omp_loop_runtime_next_real)(int*,int*) = NULL;

//static int (*GOMP_loop_dynamic_next_real)(long*,long*) = NULL;
static int (*__omp_loop_dynamic_next_real)(int*,int*) = NULL;

//static int (*GOMP_loop_guided_next_real)(long*,long*) = NULL;
static int (*__omp_loop_guided_next_real)(int*,int*) = NULL;

//static int (*GOMP_loop_static_start_real)(long,long,long,long,long*,long*) = NULL;
static int (*__omp_loop_static_start_real)(int,int,int,int*,int*,int) = NULL;

//static int (*GOMP_loop_runtime_start_real)(long,long,long,long,long*,long*) = NULL;
static void (*__omp_loop_runtime_start_real)(int,int,int,int*,int*,int) = NULL;

//static int (*GOMP_loop_guided_start_real)(long,long,long,long,long*,long*) = NULL;
static int (*__omp_loop_guided_start_ptr_real)(void*,void*,int,void**,void**,int) = NULL;

//static int (*GOMP_loop_dynamic_start_real)(long,long,long,long,long*,long*) = NULL;
static int (*__omp_loop_dynamic_start_ptr_real)(void*,void*,int,void**,void**,int) = NULL;

//static void (*GOMP_loop_end_real)(void) = NULL;
static void (*__omp_loop_end_real)(void) = NULL;

//static void (*GOMP_loop_end_nowait_real)(void) = NULL;
// ???

//static unsigned (*GOMP_sections_start_real)(unsigned) = NULL;
// ??? static void (*__omp_sections_init_real)(void) = NULL;

//static unsigned (*GOMP_sections_next_real)(void) = NULL;
// ???
//static void (*GOMP_sections_end_real)(void) = NULL;
// ???
//static void (*GOMP_sections_end_nowait_real)(void) = NULL;
// ???
//static void (*GOMP_parallel_sections_start_real)(void*,void*,unsigned,unsigned) = NULL;
// ???

//static void (*GOMP_task_real)(void*,void*,void*,long,long,int,unsigned) = NULL;
static void (*__omp_create_task_real)(void*,void*,int,int,int) = NULL;

//static void (*GOMP_taskwait_real)(void) = NULL;
static void (*__omp_taskwait_real)(void) = NULL;

//static int (*GOMP_loop_ordered_static_start_real)(long, long, long, long, long *, long *) = NULL;
static void (*__omp_loop_static_ordered_start_real)(int,int,int,int*,int*,int) = NULL;

//static int (*GOMP_loop_ordered_runtime_start_real)(long, long, long, long, long *, long *) = NULL;
static void (*__omp_loop_runtime_ordered_start_real)(int,int,int,int*,int*,int) = NULL;

//static int (*GOMP_loop_ordered_dynamic_start_real)(long, long, long, long, long *, long *) = NULL;
static void (*__omp_loop_dynamic_ordered_start_real)(void*,void*,int,void**,void**,int) = NULL;

//static int (*GOMP_loop_ordered_guided_start_real)(long, long, long, long, long *, long *) = NULL;
static void (*__omp_loop_guided_ordered_start_real)(int,int,int,int*,int*,int) = NULL;

#define INC_IF_NOT_NULL(ptr,cnt) (cnt = (ptr == NULL)?cnt:cnt+1)

static int lcc_libomp_1_0_GetOpenMPHookPoints (int rank)
{
	int count = 0;

	UNREFERENCED_PARAMETER(rank);

	/* Obtain @ for __omp_parallel_start */
	__omp_parallel_start_real =
		(void(*)(void*,void*)) dlsym (RTLD_NEXT, "__omp_parallel_start");
	INC_IF_NOT_NULL(__omp_parallel_start_real,count);

	/* Obtain @ for __omp_parallel_end */
	__omp_parallel_end_real =
		(void(*)(void)) dlsym (RTLD_NEXT, "__omp_parallel_end");
	INC_IF_NOT_NULL(__omp_parallel_end_real,count);

	/* Obtain @ for __omp_barrier */
	__omp_barrier_real =
		(void(*)(void)) dlsym (RTLD_NEXT, "__omp_barrier");
	INC_IF_NOT_NULL(__omp_barrier_real,count);

	/* Obtain @ for __omp_atomic_enter */
	__omp_atomic_enter_real =
		(void(*)(void)) dlsym (RTLD_NEXT, "__omp_atomic_enter");
	INC_IF_NOT_NULL(__omp_atomic_enter_real,count);

	/* Obtain @ for __omp_atomic_exit */
	__omp_atomic_exit_real =
		(void(*)(void)) dlsym (RTLD_NEXT, "__omp_atomic_exit");
	INC_IF_NOT_NULL(__omp_atomic_exit_real,count);

	/* Obtain @ for __omp_critical_enter */
	__omp_critical_enter_real =
		(void(*)(void)) dlsym (RTLD_NEXT, "__omp_critical_enter");
	INC_IF_NOT_NULL(__omp_critical_enter_real,count);

	/* Obtain @ for __omp_critical_exit */
	__omp_critical_exit_real =
		(void(*)(void)) dlsym (RTLD_NEXT, "__omp_critical_exit");
	INC_IF_NOT_NULL(__omp_critical_exit_real,count);

	/* Obtain @ for __omp_critical_name_enter */
	__omp_critical_name_enter_real =
		(void(*)(void*)) dlsym (RTLD_NEXT, "__omp_critical_name_enter");
	INC_IF_NOT_NULL(__omp_critical_name_enter_real,count);

	/* Obtain @ for __omp_critical_name_exit */
	__omp_critical_name_exit_real =
		(void(*)(void*)) dlsym (RTLD_NEXT, "__omp_critical_name_exit");
	INC_IF_NOT_NULL(__omp_critical_name_exit_real,count);

	/* Obtain @ for __omp_loop_static_next */
	__omp_loop_static_next_real =
		(int(*)(int*,int*)) dlsym (RTLD_NEXT, "__omp_loop_static_next");
	INC_IF_NOT_NULL(__omp_loop_static_next_real,count);

	/* Obtain @ for __omp_loop_runtime_next */
	__omp_loop_runtime_next_real =
		(void(*)(int*,int*)) dlsym (RTLD_NEXT, "__omp_loop_runtime_next");
	INC_IF_NOT_NULL(__omp_loop_runtime_next_real,count);

	/* Obtain @ for __omp_loop_guided_next */
	__omp_loop_guided_next_real =
		(int(*)(int*,int*)) dlsym (RTLD_NEXT, "__omp_loop_guided_next");
	INC_IF_NOT_NULL(__omp_loop_guided_next_real,count);

	/* Obtain @ for __omp_loop_dynamic_next */
	__omp_loop_dynamic_next_real =
		(int(*)(int*,int*)) dlsym (RTLD_NEXT, "__omp_loop_dynamic_next");
	INC_IF_NOT_NULL(__omp_loop_dynamic_next_real,count);

	/* Obtain @ for __omp_loop_static_start */
	__omp_loop_static_start_real =
		(int(*)(int,int,int,int*,int*,int)) dlsym (RTLD_NEXT, "__omp_loop_static_start");
	INC_IF_NOT_NULL(__omp_loop_static_start_real,count);

	/* Obtain @ for __omp_loop_runtime_start */
	__omp_loop_runtime_start_real =
		(void(*)(int,int,int,int*,int*,int)) dlsym (RTLD_NEXT, "__omp_loop_runtime_start");
	INC_IF_NOT_NULL(__omp_loop_runtime_start_real,count);

	/* Obtain @ for __omp_loop_guided_start_ptr */
	__omp_loop_guided_start_ptr_real =
		(int(*)(void*,void*,int,void**,void**,int)) dlsym (RTLD_NEXT, "__omp_loop_guided_start_ptr");
	INC_IF_NOT_NULL(__omp_loop_guided_start_ptr_real,count);

	/* Obtain @ for __omp_loop_dynamic_start_ptr */
	__omp_loop_dynamic_start_ptr_real =
		(int(*)(void*,void*,int,void**,void**,int)) dlsym (RTLD_NEXT, "__omp_loop_dynamic_start_ptr");
	INC_IF_NOT_NULL(__omp_loop_dynamic_start_ptr_real,count);

	/* Obtain @ for __omp_loop_end */
	__omp_loop_end_real =
		(void(*)(void)) dlsym (RTLD_NEXT, "__omp_loop_end");
	INC_IF_NOT_NULL(__omp_loop_end_real,count);

	/* Obtain @ for __omp_create_task */
	__omp_create_task_real =
		(void(*)(void*,void*,int,int,int)) dlsym (RTLD_NEXT, "__omp_create_task");
	INC_IF_NOT_NULL(__omp_create_task_real,count);

	/* Obtain @ for __omp_taskwait */
	__omp_taskwait_real = (void(*)(void)) dlsym (RTLD_NEXT, "__omp_taskwait");
	INC_IF_NOT_NULL(__omp_taskwait_real,count);

	/* Obtain @ for __omp_loop_static_ordered_start */
	__omp_loop_static_ordered_start_real =
		(void(*)(int,int,int,int*,int *,int)) dlsym (RTLD_NEXT, "__omp_loop_static_ordered_start");
	INC_IF_NOT_NULL(__omp_loop_static_ordered_start_real, count);

	/* Obtain @ for __omp_loop_runtime_ordered_start */
	__omp_loop_runtime_ordered_start_real =
		(void(*)(int,int,int,int*,int *,int)) dlsym (RTLD_NEXT, "__omp_loop_runtime_ordered_start");
	INC_IF_NOT_NULL(__omp_loop_runtime_ordered_start_real, count);

	/* Obtain @ for __omp_loop_dynamic_ordered_start */
	__omp_loop_dynamic_ordered_start_real =
		(void(*)(void*,void*,int,void**,void**,int)) dlsym (RTLD_NEXT, "__omp_loop_dynamic_ordered_start");
	INC_IF_NOT_NULL(__omp_loop_dynamic_ordered_start_real, count);

	/* Obtain @ for __omp_loop_guided_ordered_start */
	__omp_loop_guided_ordered_start_real =
		(void(*)(int,int,int,int*,int*,int)) dlsym (RTLD_NEXT, "__omp_loop_guided_ordered_start");
	INC_IF_NOT_NULL(__omp_loop_guided_ordered_start_real, count);

	/* Any hook point? */
	return count > 0;
}

/*
	INJECTED CODE -- INJECTED CODE -- INJECTED CODE -- INJECTED CODE
	INJECTED CODE -- INJECTED CODE -- INJECTED CODE -- INJECTED CODE
*/

static void callme_task (void *helper_ptr)
{
	struct openmp_task_st *helper = (* (struct openmp_task_st**)helper_ptr);
	void (*task_uf)(void*) = (void(*)(void*)) helper->p1;

	if (helper != NULL)
	{
		Extrae_OpenMP_TaskUF_Entry (helper->p1);
		Extrae_OpenMP_TaskID (helper->task_ctr);

		task_uf (helper->p2);
		free(helper);

		Extrae_OpenMP_Notify_NewExecutedTask();
		Extrae_OpenMP_TaskUF_Exit ();
	}
}

static volatile long long __omp_create_task_ctr = 1;
#if !defined(HAVE__SYNC_FETCH_AND_ADD)
static pthread_mutex_t __omp_create_task_ctr_mtx = PTHREAD_MUTEX_INITIALIZER;
#endif
void __omp_create_task (void* p1, void* p2, int p3, int p4, int p5)
{

#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_create_task is at %p\n", THREADID, __omp_create_task_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_create_task params %p %p %i %i %i", THREADID, p1, p2, p3, p4, p5);
#endif

	if (__omp_create_task_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Task_Entry (p1);
		Extrae_OpenMP_Notify_NewInstantiatedTask();

		struct openmp_task_st *helper = NULL;
		helper = (struct openmp_task_st *) malloc(sizeof(struct openmp_task_st));
		helper->p1 = p1;
		helper->p2 = p2;

		char * buf =  malloc(sizeof(char) * p3);
		memcpy (buf, p2, p3);
		helper->p2 = buf;

#if defined(HAVE__SYNC_FETCH_AND_ADD)
		helper->task_ctr = __sync_fetch_and_add(&__omp_create_task_ctr,1);
#else
		pthread_mutex_lock (&__omp_create_task_ctr_mtx);
		helper->task_ctr = __omp_create_task_ctr++;
		pthread_mutex_unlock (&__omp_create_task_ctr_mtx);
#endif
		Extrae_OpenMP_TaskID (helper->task_ctr);
		__omp_create_task_real (callme_task, &helper, sizeof(helper), p4, p5);
		Extrae_OpenMP_Task_Exit ();
	}
	else if (__omp_create_task_real != NULL)
	{
		__omp_create_task_real (p1, p2, p3, p4, p5);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_create_task is not hooked! Exiting!!\n");
		exit (0);
	}
}

void __omp_taskwait (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_taskwait is at %p\n", THREADID, __omp_taskwait_real);
#endif

	if (__omp_taskwait_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Taskwait_Entry();
		Extrae_OpenMP_EmitTaskStatistics();
		__omp_taskwait_real ();
		Extrae_OpenMP_Taskwait_Exit();
		Extrae_OpenMP_EmitTaskStatistics();
	}
	else if (__omp_taskwait_real != NULL)
	{
		__omp_taskwait_real ();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_taskwait is not hooked! Exiting!!\n");
		exit (0);
	}
}

void __omp_loop_end (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_end is at %p\n", THREADID, __omp_loop_end_real);
#endif

	if (__omp_loop_end_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Join_Wait_Entry();
		__omp_loop_end_real();
		Extrae_OpenMP_Join_Wait_Exit();
		Extrae_OpenMP_UF_Exit ();
		Extrae_OpenMP_DO_Exit ();
	}
	else if (__omp_loop_end_real != NULL)
	{
		__omp_loop_end_real();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_end is not hooked! exiting!!\n");
		exit (0);
	}
}

int __omp_loop_static_start (int p1, int p2, int p3, int *p4, int *p5, int p6)
{
	int res = 0;

#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_static_start is at %p\n", THREADID, __omp_loop_static_start_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %i %i %i %p %p %i\n", THREADID, p1, p2, p3, p4, p5, p6);
#endif

	if (__omp_loop_static_start_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		res = __omp_loop_static_start_real (p1, p2, p3, p4, p5, p6);
		Extrae_OpenMP_UF_Entry (par_uf);
		Backend_Leave_Instrumentation();
	}
	else if (__omp_loop_static_start_real != NULL)
	{
		res = __omp_loop_static_start_real (p1, p2, p3, p4, p5, p6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_static_start is not hooked! exiting!!\n");
		exit (0);
	}
	return res;
}

void __omp_loop_runtime_start (int p1, int p2, int p3, int *p4, int *p5, int p6)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_runtime_start is at %p\n", THREADID, __omp_loop_runtime_start_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %i %i %i %p %p %i\n", THREADID, p1, p2, p3, p4, p5, p6);
#endif

	if (__omp_loop_runtime_start_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		__omp_loop_runtime_start_real (p1, p2, p3, p4, p5, p6);
		Extrae_OpenMP_UF_Entry (par_uf);
		Backend_Leave_Instrumentation();
	}
	else if (__omp_loop_runtime_start_real != NULL)
	{
		__omp_loop_runtime_start_real (p1, p2, p3, p4, p5, p6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_runtime_start is not hooked! exiting!!\n");
		exit (0);
	}
}

int __omp_loop_guided_start_ptr (void* p1, void* p2, int p3, void** p4, void** p5 ,int p6)
{
	int res = 0;
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_guided_start_ptr is at %p\n", THREADID, __omp_loop_guided_start_ptr_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %p %p %i %p %p %i\n", THREADID, p1, p2, p3, p4, p5, p6);
#endif

	if (__omp_loop_guided_start_ptr_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		res = __omp_loop_guided_start_ptr_real (p1, p2, p3, p4, p5, p6);
		Extrae_OpenMP_UF_Entry (par_uf);
		Backend_Leave_Instrumentation();
	}
	else if (__omp_loop_guided_start_ptr_real != NULL)
	{
		res = __omp_loop_guided_start_ptr_real (p1, p2, p3, p4, p5, p6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_guided_start_ptr is not hooked! exiting!!\n");
		exit (0);
	}
	return res;
}

int __omp_loop_dynamic_start_ptr (void* p1, void* p2, int p3, void** p4, void** p5, int p6)
{
	int res = 0;
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_dynamic_start_ptr is at %p\n", THREADID, __omp_loop_dynamic_start_ptr_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %p %p %i %p %p %i\n", THREADID, p1, p2, p3, p4, p5, p6);
#endif

	if (__omp_loop_dynamic_start_ptr_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		res = __omp_loop_dynamic_start_ptr_real (p1, p2, p3, p4, p5, p6);
		Extrae_OpenMP_UF_Entry (par_uf);
		Backend_Leave_Instrumentation();
	}
	else if (__omp_loop_dynamic_start_ptr_real != NULL)
	{
		res = __omp_loop_dynamic_start_ptr_real (p1, p2, p3, p4, p5, p6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_dynamic_start_ptr is not hooked! exiting!!\n");
		exit (0);
	}
	return res;
}

int __omp_loop_static_next (int *p1, int *p2)
{
	int res = 0;
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_static_next is at %p\n", THREADID, __omp_loop_static_next_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %p %p\n", THREADID, p1, p2);
#endif

	if (__omp_loop_static_next_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Work_Entry();
		res = __omp_loop_static_next_real (p1, p2);
		Extrae_OpenMP_Work_Exit();
	}
	else if (__omp_loop_static_next_real != NULL)
	{
		res = __omp_loop_static_next_real (p1, p2);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_static_next is not hooked! exiting!!\n");
		exit (0);
	}
	return res;
}

void __omp_loop_runtime_next (int *p1, int *p2)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_runtime_next is at %p\n", THREADID, __omp_loop_runtime_next_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %p %p\n", THREADID, p1, p2);
#endif

	if (__omp_loop_runtime_next_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Work_Entry();
		__omp_loop_runtime_next_real (p1, p2);
		Extrae_OpenMP_Work_Exit();
	}
	else if (__omp_loop_runtime_next_real != NULL)
	{
		__omp_loop_runtime_next_real (p1, p2);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_runtime_next is not hooked! exiting!!\n");
		exit (0);
	}
}

int __omp_loop_guided_next (int *p1, int *p2)
{
	int res = 0;
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_guided_next is at %p\n", THREADID, __omp_loop_guided_next_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %p %p\n", THREADID, p1, p2);
#endif

	if (__omp_loop_guided_next_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Work_Entry();
		res = __omp_loop_guided_next_real (p1, p2);
		Extrae_OpenMP_Work_Exit();
	}
	else if (__omp_loop_guided_next_real != NULL)
	{
		res = __omp_loop_guided_next_real (p1, p2);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_guided_next is not hooked! exiting!!\n");
		exit (0);
	}
	return res;
}

int __omp_loop_dynamic_next (int *p1, int *p2)
{
	int res = 0;
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_dynamic_next is at %p\n", THREADID, __omp_loop_dynamic_next_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %p %p\n", THREADID, p1, p2);
#endif

	if (__omp_loop_dynamic_next_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Work_Entry();
		res = __omp_loop_dynamic_next_real (p1, p2);
		Extrae_OpenMP_Work_Exit();
	}
	else if (__omp_loop_dynamic_next_real != NULL)
	{
		res = __omp_loop_dynamic_next_real (p1, p2);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_dynamic_next is not hooked! exiting!!\n");
		exit (0);
	}
	return res;
}

void __omp_parallel_start (void *p1, void *p2)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_parallel_start is at %p\n", THREADID, __omp_parallel_start_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_parallel_start params %p %p\n", THREADID, p1, p2);
#endif

	if (__omp_parallel_start_real != NULL && EXTRAE_INITIALIZED())
	{
		/* Set the pointer to the correct PARALLEL user function */
		par_uf = (void(*)(void*))p1;

		Extrae_OpenMP_ParRegion_Entry();
		Extrae_OpenMP_EmitTaskStatistics();

		__omp_parallel_start_real (callme_par, p2);

		/* GCC/libgomp does not execute callme_par per root thread, emit
		   the required event here - call Backend to get a new time! */
		Extrae_OpenMP_UF_Entry (p1);
	}
	else if (__omp_parallel_start_real != NULL)
	{
		__omp_parallel_start_real (p1, p2);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_parallel_start is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_parallel_end (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_parallel_end is at %p\n", THREADID, __omp_parallel_end_real);
#endif

	if (__omp_parallel_start_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_UF_Exit ();
		__omp_parallel_end_real ();
		Extrae_OpenMP_ParRegion_Exit();
		Extrae_OpenMP_EmitTaskStatistics();
	}
	else if (__omp_parallel_start_real != NULL)
	{
		__omp_parallel_end_real ();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_parallel_end is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_barrier (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_barrier is at %p\n", THREADID, __omp_barrier_real);
#endif

	if (__omp_barrier_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Barrier_Entry ();
		__omp_barrier_real ();
		Extrae_OpenMP_Barrier_Exit ();
	}
	else if (__omp_barrier_real != NULL)
	{
		__omp_barrier_real ();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_barrier is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_critical_name_enter (void *p1)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_critical_name_enter is at %p\n", THREADID, __omp_critical_name_enter_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_critical_name_enter params %p\n", THREADID, p1);
#endif

	if (__omp_critical_name_enter_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Named_Lock_Entry();
		__omp_critical_name_enter_real (p1);
		Extrae_OpenMP_Named_Lock_Exit(p1);
	}
	else if (__omp_critical_name_enter_real)
	{
		__omp_critical_name_enter_real (p1);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_critical_name_enter is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_critical_name_exit (void *p1)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_critical_name_exit is at %p\n", THREADID, __omp_critical_name_exit_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_critical_name_exit params %p\n", THREADID, p1);
#endif

	if (__omp_critical_name_exit_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Named_Unlock_Entry(p1);
		__omp_critical_name_exit_real (p1);
		Extrae_OpenMP_Named_Unlock_Exit();
	}
	else if (__omp_critical_name_exit_real != NULL)
	{
		__omp_critical_name_exit_real (p1);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_critical_name_exit is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_critical_enter (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_critical_enter is at %p\n", THREADID, __omp_critical_enter_real);
#endif

	if (__omp_critical_enter_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Unnamed_Lock_Entry();
		__omp_critical_enter_real();
		Extrae_OpenMP_Unnamed_Lock_Exit();
	}
	else if (__omp_critical_enter_real != NULL)
	{
		__omp_critical_enter_real();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_critical_enter is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_critical_exit (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_critical_exit is at %p\n", THREADID, __omp_critical_exit_real);
#endif

	if (__omp_critical_exit_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Unnamed_Unlock_Entry();
		__omp_critical_exit_real ();
		Extrae_OpenMP_Unnamed_Unlock_Exit();
	}
	else if (__omp_critical_exit_real != NULL)
	{
		__omp_critical_exit_real ();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_critical_exit is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_atomic_enter (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_atomic_enter is at %p\n", THREADID, __omp_atomic_enter_real);
#endif

	if (__omp_atomic_enter_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Unnamed_Lock_Entry();
		__omp_atomic_enter_real();
		Extrae_OpenMP_Unnamed_Lock_Exit();
	}
	else if (__omp_atomic_enter_real != NULL)
	{
		__omp_atomic_enter_real();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_atomic_enter is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_atomic_exit (void)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_atomic_exit is at %p\n", THREADID, __omp_atomic_exit_real);
#endif

	if (__omp_atomic_exit_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_Unnamed_Unlock_Entry();
		__omp_atomic_exit_real ();
		Extrae_OpenMP_Unnamed_Unlock_Exit();
	}
	else if (__omp_atomic_exit_real != NULL)
	{
		__omp_atomic_exit_real ();
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_atomic_exit is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_loop_static_ordered_start (int P1, int P2, int P3, int *P4, int *P5, int P6)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_static_ordered_start %p\n", THREADID, __omp_loop_static_ordered_start_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %i %i %i %p %p %i\n", THREADID, P1, P2, P3, P4, P5, P6);
#endif

	if (__omp_loop_static_ordered_start_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		__omp_loop_static_ordered_start_real (P1, P2, P3, P4, P5, P6);
		Extrae_OpenMP_UF_Entry (par_uf);
	}
	else if (__omp_loop_static_ordered_start_real != NULL)
	{
		__omp_loop_static_ordered_start_real (P1, P2, P3, P4, P5, P6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_static_ordered_start_real is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_loop_runtime_ordered_start (int P1, int P2, int P3, int *P4, int *P5, int P6)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_runtime_ordered_start %p\n", THREADID, __omp_loop_runtime_ordered_start_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %i %i %i %p %p %i\n", THREADID, P1, P2, P3, P4, P5, P6);
#endif

	if (__omp_loop_runtime_ordered_start_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		__omp_loop_runtime_ordered_start_real (P1, P2, P3, P4, P5, P6);
		Extrae_OpenMP_UF_Entry (par_uf);
	}
	else if (__omp_loop_runtime_ordered_start_real != NULL)
	{
		__omp_loop_runtime_ordered_start_real (P1, P2, P3, P4, P5, P6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_runtime_ordered_start_real is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_loop_dynamic_ordered_start (void *P1, void *P2, int P3, void **P4, void **P5, int P6)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_dynamic_ordered_start %p\n", THREADID, __omp_loop_dynamic_ordered_start_real);
    fprintf (stderr, PACKAGE_NAME": THREAD %d params %p %p %i %p %p %i\n", THREADID, P1, P2, P3, P4, P5, P6);
#endif

	if (__omp_loop_dynamic_ordered_start_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		__omp_loop_dynamic_ordered_start_real (P1, P2, P3, P4, P5, P6);
		Extrae_OpenMP_UF_Entry (par_uf);
	}
	else if (__omp_loop_dynamic_ordered_start_real != NULL)
	{
		__omp_loop_dynamic_ordered_start_real (P1, P2, P3, P4, P5, P6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_dynamic_ordered_start_real is not hooked! exiting!!\n");
		exit (0);
	}
}

void __omp_loop_guided_ordered_start (int P1, int P2, int P3, int *P4, int *P5, int P6)
{
#if defined(DEBUG)
	fprintf (stderr, PACKAGE_NAME": THREAD %d __omp_loop_guided_ordered_start %p\n", THREADID, __omp_loop_guided_ordered_start_real);
	fprintf (stderr, PACKAGE_NAME": THREAD %d params %i %i %i %p %p %i\n", THREADID, P1, P2, P3, P4, P5, P6);
#endif

	if (__omp_loop_guided_ordered_start_real != NULL && EXTRAE_INITIALIZED())
	{
		Extrae_OpenMP_DO_Entry ();
		__omp_loop_guided_ordered_start_real (P1, P2, P3, P4, P5, P6);
		Extrae_OpenMP_UF_Entry (par_uf);
	}
	else if (__omp_loop_guided_ordered_start_real != NULL)
	{
		__omp_loop_guided_ordered_start_real (P1, P2, P3, P4, P5, P6);
	}
	else
	{
		fprintf (stderr, PACKAGE_NAME": __omp_loop_guided_ordered_start_real is not hooked! exiting!!\n");
		exit (0);
	}
}

extern int omp_get_max_threads();

int lcc_libomp_1_0_hook_points (int ntask)
{
	return lcc_libomp_1_0_GetOpenMPHookPoints (ntask);
}

#endif /* PIC */
