#!/usr/bin/env perl
# Copyright (c) 2020-2024 AO MCST
# Distributed under the terms of MIT License.

# Usage: perl papi_monitors.pl <version> <filename>
#
# Where version - commands system version:
# 	1: MCST elbrus-v1 (e3m) 
# 	2: MCST elbrus-v2 (e3s) 
# 	3: MCST elbrus-v3 (e2s) 
# 	4: MCST elbrus-v4 
# 	5: MCST elbrus-v5
# 	6: MCST elbrus-v6
#
# filename - file with information about events

use strict;
use warnings;
use File::Basename;

my ($version, $filename) = @ARGV;
print "/* Generated by " . basename($0) . " */\n";
printf "static const e2k_entry_t v%s_pe[] = {\n", $version;

sub split_text {
    my ($text, $len) = @_;
    my @result = ();

    while (length $text > $len) {
        my $i;
        for ($i = $len-1; $i > 0 and substr($text, $i, 1) !~ /\s/; $i--) {
        }
        $i = $len if $i == 0;
        push @result, substr($text, 0, $i);
        $text = substr $text, $i;
        $text =~ s/^\s+//;
    }

    push @result, $text;
    return @result;
}
open (FILE,'<',$filename); 
while (<FILE>) {
    if (/^#/) {
        chomp;
        print "/* $_ */\n";
    }
# e2k_monitors file format
# mask channel:id	SHORT_NAME | Description
# mask is bit mask of allowed channels: es0 (DDMAR0), es1 (DDMAR1), DIMAR0, DIMAR1
    elsif (/^(\d+) \d:([[:xdigit:]]+)\s+(\S+) \| (.+)/) {
        my $mask = oct("0b".reverse($1));
        my $num = $2;
        my $name = $3;
        my $descr = $4;
        $descr =~ s/\"/\\\"/g;
        $_ = sprintf "\t{\t.name = \"%s\",\n\t\t.desc = \"%s\",\n\t\t.mask =  0x%x,\n\t\t.code =  0x%02s,\n\t}, \n", $name, $descr, $mask, $num;
        print;
    }
}
print "};\n";
printf "#define PME_E2K_V%s_EVENT_COUNT       (sizeof(v%s_pe)/sizeof(e2k_entry_t))\n", $version, $version; 

