/* Copyright (c) 2020-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#include "pfmlib_priv.h"
#include "pfmlib_e2k_priv.h"
#include "../include/perfmon/pfmlib.h"
#include "events/e2k_elbrus_v1_events.h"
#include "events/e2k_elbrus_v2_events.h"
#include "events/e2k_elbrus_v3_events.h"
#include "events/e2k_elbrus_v4_events.h"
#include "events/e2k_elbrus_v5_events.h"
#include "events/e2k_elbrus_v6_events.h"

pfmlib_pmu_t e2k_elbrus_v1_support={
        .desc                   = "MCST elbrus (e3m)",
        .name                   = "elbrus-v1",
        .pmu                    = PFM_PMU_E2K_V1,
        .pme_count              = LIBPFM_ARRAY_SIZE(v1_pe),
        .type                   = PFM_PMU_TYPE_CORE,
        .supported_plm          = E2K_PLM,
        .max_encoding           = 4,
        .num_cntrs              = 4,
        .pe                     = v1_pe,
        .atdesc                 = NULL,
        .flags                  = 0,

        .pmu_detect             = pfm_e2k_detect,
        .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding,
         PFMLIB_ENCODE_PERF(pfm_e2k_get_perf_encoding),
        .get_event_first        = pfm_e2k_get_event_first,
        .get_event_next         = pfm_e2k_get_event_next,
        .event_is_valid         = pfm_e2k_event_is_valid,
        .validate_table         = pfm_e2k_validate_table,
        .get_event_info         = pfm_e2k_get_event_info,
        .get_event_attr_info    = pfm_e2k_get_event_attr_info,
         PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs),
        .get_event_nattrs       = pfm_e2k_get_event_nattrs,
};

pfmlib_pmu_t e2k_elbrus_v2_support={
        .desc                   = "MCST elbrus-2c+",
        .name                   = "elbrus-v2",
        .pmu                    = PFM_PMU_E2K_V2,
        .pme_count              = LIBPFM_ARRAY_SIZE(v2_pe),
        .type                   = PFM_PMU_TYPE_CORE,
        .supported_plm          = E2K_PLM,
        .max_encoding           = 4,
        .num_cntrs              = 4,
        .pe                     = v2_pe,
        .atdesc                 = NULL,
        .flags                  = 0,

        .pmu_detect             = pfm_e2k_detect,
        .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding,
         PFMLIB_ENCODE_PERF(pfm_e2k_get_perf_encoding),
        .get_event_first        = pfm_e2k_get_event_first,
        .get_event_next         = pfm_e2k_get_event_next,
        .event_is_valid         = pfm_e2k_event_is_valid,
        .validate_table         = pfm_e2k_validate_table,
        .get_event_info         = pfm_e2k_get_event_info,
        .get_event_attr_info    = pfm_e2k_get_event_attr_info,
         PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs),
        .get_event_nattrs       = pfm_e2k_get_event_nattrs,
};

pfmlib_pmu_t e2k_elbrus_v3_support={
        .desc                   = "MCST elbrus-4c (e2s)",
        .name                   = "elbrus-v3",
        .pmu                    = PFM_PMU_E2K_V3,
        .pme_count              = LIBPFM_ARRAY_SIZE(v3_pe),
        .type                   = PFM_PMU_TYPE_CORE,
        .supported_plm          = E2K_PLM,
        .max_encoding           = 4,
        .num_cntrs              = 4,
        .pe                     = v3_pe,
        .atdesc                 = NULL,
        .flags                  = 0,

        .pmu_detect             = pfm_e2k_detect,
        .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding,
         PFMLIB_ENCODE_PERF(pfm_e2k_get_perf_encoding),
        .get_event_first        = pfm_e2k_get_event_first,
        .get_event_next         = pfm_e2k_get_event_next,
        .event_is_valid         = pfm_e2k_event_is_valid,
        .validate_table         = pfm_e2k_validate_table,
        .get_event_info         = pfm_e2k_get_event_info,
        .get_event_attr_info    = pfm_e2k_get_event_attr_info,
         PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs),
        .get_event_nattrs       = pfm_e2k_get_event_nattrs,
};

pfmlib_pmu_t e2k_elbrus_v4_support={
        .desc                   = "MCST elbrus-8c",
        .name                   = "elbrus-v4",
        .pmu                    = PFM_PMU_E2K_V4,
        .pme_count              = LIBPFM_ARRAY_SIZE(v4_pe),
        .type                   = PFM_PMU_TYPE_CORE,
        .supported_plm          = E2K_PLM,
        .max_encoding           = 4,
        .num_cntrs              = 4,
        .pe                     = v4_pe,
        .atdesc                 = NULL,
        .flags                  = 0,

        .pmu_detect             = pfm_e2k_detect,
        .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding,
         PFMLIB_ENCODE_PERF(pfm_e2k_get_perf_encoding),
        .get_event_first        = pfm_e2k_get_event_first,
        .get_event_next         = pfm_e2k_get_event_next,
        .event_is_valid         = pfm_e2k_event_is_valid,
        .validate_table         = pfm_e2k_validate_table,
        .get_event_info         = pfm_e2k_get_event_info,
        .get_event_attr_info    = pfm_e2k_get_event_attr_info,
         PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs),
        .get_event_nattrs       = pfm_e2k_get_event_nattrs,
};

pfmlib_pmu_t e2k_elbrus_v5_support={
        .desc                   = "MCST elbrus-8c2",
        .name                   = "elbrus-v5",
        .pmu                    = PFM_PMU_E2K_V5,
        .pme_count              = LIBPFM_ARRAY_SIZE(v5_pe),
        .type                   = PFM_PMU_TYPE_CORE,
        .supported_plm          = E2K_PLM,
        .max_encoding           = 4,
        .num_cntrs              = 4,
        .pe                     = v5_pe,
        .atdesc                 = NULL,
        .flags                  = 0,

        .pmu_detect             = pfm_e2k_detect,
        .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding,
         PFMLIB_ENCODE_PERF(pfm_e2k_get_perf_encoding),
        .get_event_first        = pfm_e2k_get_event_first,
        .get_event_next         = pfm_e2k_get_event_next,
        .event_is_valid         = pfm_e2k_event_is_valid,
        .validate_table         = pfm_e2k_validate_table,
        .get_event_info         = pfm_e2k_get_event_info,
        .get_event_attr_info    = pfm_e2k_get_event_attr_info,
         PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs),
        .get_event_nattrs       = pfm_e2k_get_event_nattrs,
};

pfmlib_pmu_t e2k_elbrus_v6_support={
        .desc                   = "MCST elbrus-e16c",
        .name                   = "elbrus-v6",
        .pmu                    = PFM_PMU_E2K_V6,
        .pme_count              = LIBPFM_ARRAY_SIZE(v6_pe),
        .type                   = PFM_PMU_TYPE_CORE,
        .supported_plm          = E2K_PLM,
        .max_encoding           = 4,
        .num_cntrs              = 4,
        .pe                     = v6_pe,
        .atdesc                 = NULL,
        .flags                  = 0,

        .pmu_detect             = pfm_e2k_detect,
        .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding,
         PFMLIB_ENCODE_PERF(pfm_e2k_get_perf_encoding),
        .get_event_first        = pfm_e2k_get_event_first,
        .get_event_next         = pfm_e2k_get_event_next,
        .event_is_valid         = pfm_e2k_event_is_valid,
        .validate_table         = pfm_e2k_validate_table,
        .get_event_info         = pfm_e2k_get_event_info,
        .get_event_attr_info    = pfm_e2k_get_event_attr_info,
         PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs),
        .get_event_nattrs       = pfm_e2k_get_event_nattrs,
};

