/* Copyright (c) 2020-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>


#include "pfmlib_priv.h"
#include "pfmlib_e2k_priv.h"
#include "pfmlib_perf_event_priv.h"

int
pfm_e2k_get_perf_encoding(void *this, pfmlib_event_desc_t *e)
{
	struct perf_event_attr *attr = e->os_data;
	int ret;

	ret = pfm_e2k_get_encoding(this, e);
	if (ret != PFM_SUCCESS)
		return ret;

	attr->type = PERF_TYPE_RAW;
	attr->config = (e->codes[1] << 8) | e->codes[0];
	return PFM_SUCCESS;
}

void
pfm_e2k_perf_validate_pattrs(void *this, pfmlib_event_desc_t *e)
{
	int i, compact;

	for (i = 0; i < e->npattrs; i++) {
		compact = 0;

		if (e->pattrs[i].type == PFM_ATTR_UMASK)
			continue;

		if (e->pattrs[i].ctrl == PFM_ATTR_CTRL_PMU) {
			if (  e->pattrs[i].idx == E2K_ATTR_K
			    ||e->pattrs[i].idx == E2K_ATTR_U)
				compact = 1;
		}
		if (e->pattrs[i].ctrl == PFM_ATTR_CTRL_PERF_EVENT) {
			if (e->pattrs[i].idx == PERF_ATTR_PR)
				compact = 1;
		}

		if (compact) {
			pfmlib_compact_pattrs(e, i);
			i--;
		}
	}
}

static int
find_pmu_type_by_name(const char *name)
{
    char filename[PATH_MAX];
    FILE *fp;
    int ret, type;

    if (!name)
    	return PFM_ERR_NOTSUPP;
    
    sprintf(filename, "/sys/bus/event_source/devices/%s/type", name);
    fp = fopen(filename, "r");
    if (!fp)
    	return PFM_ERR_NOTSUPP;

    ret = fscanf(fp, "%d", &type);
    if (ret != 1)
    	type = PFM_ERR_NOTSUPP;
	
    fclose(fp);

    return type;
}


int
pfm_e2k_uncore_get_perf_encoding(void *this, pfmlib_event_desc_t *e)
{
    pfmlib_pmu_t *pmu = this;
    struct perf_event_attr *attr = e->os_data;
    int ret;
    
    if (!pmu->get_event_encoding[PFM_OS_NONE])
            return PFM_ERR_NOTSUPP;

    ret = pmu->get_event_encoding[PFM_OS_NONE](this, e);
    if (ret != PFM_SUCCESS)
   	 return ret;
    
    ret = find_pmu_type_by_name(pmu->perf_name);
    if (ret < 0)
    	return ret;
    
    attr->type = ret;
    
    attr->config = e->codes[0];
    
    return PFM_SUCCESS;
}

