/* Copyright (c) 2020-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#include "pfmlib_priv.h"
#include "pfmlib_e2k_priv.h"
#include "../include/perfmon/pfmlib.h"
#include "events/e2k_elbrus_uncore_events.h"

static int pfm_e2k_uncore_detect(void* this)
{
    if((__builtin_cpu_is ("elbrus-v3")) || (__builtin_cpu_is ("elbrus-v4")) || (__builtin_cpu_is ("elbrus-v5")) || (__builtin_cpu_is ("elbrus-v6")))
            return PFM_SUCCESS;
    return PFM_ERR_NOTSUPP;
}

#define DEFINE_SIC_MCM0_NUMA_NODE(n) \
pfmlib_pmu_t e2k_elbrus_uncore_sic_##n##_MCM0_support={ \
    .desc           = "MCST elbrus uncore sic_"#n"_MCM0", \
    .name           = "sic_"#n"_MCM0", \
    .pmu            = PFM_PMU_E2K_SIC_##n##_MCM0_UNCORE, \
    .perf_name      = "sic_"#n"_MCM0", \
    .pme_count      = LIBPFM_ARRAY_SIZE(elbrus_uncore_sic_MCM0_pe), \
    .type           = PFM_PMU_TYPE_UNCORE, \
    .num_cntrs      = 1, \
    .num_fixed_cntrs    = 0, \
    .max_encoding       = 1, \
    .pe         = elbrus_uncore_sic_MCM0_pe, \
    .pmu_detect     = pfm_e2k_uncore_detect, \
    .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding, \
    PFMLIB_ENCODE_PERF(pfm_e2k_uncore_get_perf_encoding), \
    PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs), \
    .get_event_first    = pfm_e2k_get_event_first, \
    .get_event_next     = pfm_e2k_get_event_next, \
    .event_is_valid     = pfm_e2k_event_is_valid, \
    .validate_table     = pfm_e2k_validate_table, \
    .get_event_info     = pfm_e2k_get_event_info, \
    .get_event_attr_info    = pfm_e2k_get_event_attr_info, \
};

#define DEFINE_SIC_MCM1_NUMA_NODE(n) \
pfmlib_pmu_t e2k_elbrus_uncore_sic_##n##_MCM1_support={ \
    .desc           = "MCST elbrus uncore sic_"#n"_MCM0", \
    .name           = "sic_"#n"_MCM1", \
    .pmu            = PFM_PMU_E2K_SIC_##n##_MCM1_UNCORE, \
    .perf_name      = "sic_"#n"_MCM1", \
    .pme_count      = LIBPFM_ARRAY_SIZE(elbrus_uncore_sic_MCM1_pe), \
    .type           = PFM_PMU_TYPE_UNCORE, \
    .num_cntrs      = 1, \
    .num_fixed_cntrs    = 0, \
    .max_encoding       = 1, \
    .pe         = elbrus_uncore_sic_MCM1_pe, \
    .pmu_detect     = pfm_e2k_uncore_detect, \
    .get_event_encoding[PFM_OS_NONE] = pfm_e2k_get_encoding, \
    PFMLIB_ENCODE_PERF(pfm_e2k_uncore_get_perf_encoding), \
    PFMLIB_VALID_PERF_PATTRS(pfm_e2k_perf_validate_pattrs), \
    .get_event_first    = pfm_e2k_get_event_first, \
    .get_event_next     = pfm_e2k_get_event_next, \
    .event_is_valid     = pfm_e2k_event_is_valid, \
    .validate_table     = pfm_e2k_validate_table, \
    .get_event_info     = pfm_e2k_get_event_info, \
    .get_event_attr_info    = pfm_e2k_get_event_attr_info, \
};
					
DEFINE_SIC_MCM0_NUMA_NODE(0);
DEFINE_SIC_MCM0_NUMA_NODE(1);
DEFINE_SIC_MCM0_NUMA_NODE(2);
DEFINE_SIC_MCM0_NUMA_NODE(3);

DEFINE_SIC_MCM1_NUMA_NODE(0);
DEFINE_SIC_MCM1_NUMA_NODE(1);
DEFINE_SIC_MCM1_NUMA_NODE(2);
DEFINE_SIC_MCM1_NUMA_NODE(3);

