/* Copyright (c) 2020-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>


#include "pfmlib_priv.h"
#include "pfmlib_mcst_sparc_priv.h"
#include "pfmlib_perf_event_priv.h"

int
pfm_mcst_sparc_get_perf_encoding(void *this, pfmlib_event_desc_t *e)
{
	struct perf_event_attr *attr = e->os_data;
	int ret;

	ret = pfm_mcst_sparc_get_encoding(this, e);
	if (ret != PFM_SUCCESS)
		return ret;

	attr->type = PERF_TYPE_RAW;
	attr->config = (e->codes[0] << 16) | e->codes[1];
	return PFM_SUCCESS;
}

void
pfm_mcst_sparc_perf_validate_pattrs(void *this, pfmlib_event_desc_t *e)
{
	int i, compact;

	for (i = 0; i < e->npattrs; i++) {
		compact = 0;

		if (e->pattrs[i].type == PFM_ATTR_UMASK)
			continue;

		if (e->pattrs[i].ctrl == PFM_ATTR_CTRL_PMU) {
			if (  e->pattrs[i].idx == MCST_SPARC_ATTR_K
			    ||e->pattrs[i].idx == MCST_SPARC_ATTR_U)
				compact = 1;
		}
		if (e->pattrs[i].ctrl == PFM_ATTR_CTRL_PERF_EVENT) {
			if (e->pattrs[i].idx == PERF_ATTR_PR)
				compact = 1;
		}

		if (compact) {
			pfmlib_compact_pattrs(e, i);
			i--;
		}
	}
}

