/* Copyright (c) 2020-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#ifndef __PFMLIB_MCST_SPARC_PRIV_H__
#define __PFMLIB_MCST_SPARC_H__

typedef struct {
	char                    *name;  /* event name */
	char                    *desc;  /* event description */
	int			mask;
	int                     code;   
} mcst_sparc_entry_t;

typedef union {
        unsigned int val;
	struct {
		unsigned int    ctrl_s0   : 1;
		unsigned int    ctrl_s1   : 1;
		unsigned int    ctrl_s2   : 1;
		unsigned int    ctrl_s3   : 1;
		unsigned int    reserved1 : 12;
		unsigned int    code      : 16;
	        } config;
} pfm_mcst_sparc_reg_t;

#define PME_CTRL_S0             0
#define PME_CTRL_S1             1
#define PME_CTRL_S2             2
#define PME_CTRL_S3             3

#define MCST_SPARC_ATTR_K    0
#define MCST_SPARC_ATTR_U    1
#define MCST_SPARC_NUM_ATTRS 2

#define MCST_SPARC_PLM (PFM_PLM0|PFM_PLM3)
extern int pfm_mcst_sparc_detect(void *this);
extern int pfm_mcst_sparc_get_encoding(void *this, pfmlib_event_desc_t *e);
extern int pfm_mcst_sparc_get_event_first(void *this);
extern int pfm_mcst_sparc_get_event_next(void *this, int idx);
extern int pfm_mcst_sparc_event_is_valid(void *this, int pidx);
extern int pfm_mcst_sparc_validate_table(void *this, FILE *fp);
extern int pfm_mcst_sparc_get_event_attr_info(void *this, int pidx, int attr_idx, pfm_event_attr_info_t *info);
extern int pfm_mcst_sparc_get_event_info(void *this, int idx, pfm_event_info_t *info);
extern unsigned int pfm_mcst_sparc_get_event_nattrs(void *this, int pidx);

extern void pfm_mcst_sparc_perf_validate_pattrs(void *this, pfmlib_event_desc_t *e);

extern int pfm_mcst_sparc_get_perf_encoding(void *this, pfmlib_event_desc_t *e); 

#endif /* __PFMLIB_MCST_SPARC_PRIV_H__ */
