/*
 * Copyright (c) 2020-2021 AO MCST
 * Distributed under the terms of MIT License.
 */

#include <errno.h>
#include <error.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <unistd.h>
#include <asm/e2k_syswork.h>

asm(
"       .global rb_e2k_flushr\n"
"       .type rb_e2k_flushr,@function\n"
"       .align 8\n"
"rb_e2k_flushr:\n"
"       flushr\n"
"       return %ctpr3\n"
"       {\n"
"           ct %ctpr3\n"
"           ipd 3\n"
"       }\n"
"       .size   rb_e2k_flushr, .- rb_e2k_flushr\n"
"");

typedef unsigned long long __u64;

#ifndef __NR_access_hw_stacks
#define __NR_access_hw_stacks 254
/* modes for sys_access_hw_stacks */
enum {
       E2K_READ_CHAIN_STACK,
       E2K_READ_PROCEDURE_STACK,
       E2K_WRITE_PROCEDURE_STACK,
       E2K_GET_CHAIN_STACK_OFFSET,
       E2K_GET_CHAIN_STACK_SIZE,
       E2K_GET_PROCEDURE_STACK_SIZE,
       E2K_READ_CHAIN_STACK_EX,
       E2K_READ_PROCEDURE_STACK_EX,
       E2K_WRITE_PROCEDURE_STACK_EX,
       E2K_WRITE_CHAIN_STACK_EX,
};
#endif

size_t rb_e2k_get_procedure_stack(__u64 **ps_buf)
{
    __u64 ps_size, ps_index;
    if (syscall(__NR_access_hw_stacks, E2K_GET_PROCEDURE_STACK_SIZE,
                    NULL, NULL, 0, &ps_size) == -1)
        error(1, 0, "E2K_GET_PROCEDURE_STACK_SIZE");

    *ps_buf = realloc(*ps_buf, ps_size);
    if (!*ps_buf)
        error(ENOMEM, ENOMEM, "realloc");

    ps_index = 0;
    if (syscall(__NR_access_hw_stacks, E2K_READ_PROCEDURE_STACK_EX,
            &ps_index, *ps_buf, ps_size, NULL))
        error(1, errno, "E2K_READ_PROCEDURE_STACK_EX");

    return ps_size;
}
