#ifndef __ELO_H__
#define __ELO_H__

#include "config.h"

#include <xorg-server.h>
#include <unistd.h>
#include <misc.h>
#include <xf86.h>
#include <math.h>
#include <stdio.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <exevents.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/input.h>
#include <linux/version.h>
#include <signal.h>
#include <dirent.h>
#include <getopt.h>

#include "elo_version.h"

/* String constants */
#define ELO_DRIVER_NAME         "elo"

#define ELO_DEV_INPUT_EVENT     "/dev/input"
#define ELO_EVENT_DEV_NAME      "event"

#define ELO_DEVICE_NAME         "Elo TouchSystems"

/* Common macros */
#ifndef SYSCALL
#define SYSCALL( call ) while ( ( ( call ) == -1 ) && ( errno == EINTR ) )
#endif

#ifndef WORD_ASSEMBLY
#define WORD_ASSEMBLY( byte1, byte2 ) ( ( ( byte2 ) << 8 ) | ( byte1 ) )
#endif

#define ELO_STATE_PRESS     0x01
#define ELO_STATE_STREAM    0x02
#define ELO_STATE_RELEASE   0x04

#define ELO_BUTTON_THRESHOLD_MAX   255

typedef enum {
    CONTROLLER_TYPE_USB,
    CONTROLLER_TYPE_RS232
} EloControllerType;

typedef struct _EloDeviceRec {
    char *device;     /* Device name from xorg configuration file, like a "/dev/input/mouse0" */
    int controller;   /* Controller connection type USB or RS232 */

    int screenNumber;
    int screenWidth;
    int screenHeight;

    int minX;
    int minY;
    int maxX;
    int maxY;
    int reverseX;
    int reverseY;

    int debug;
    int threshold;

    int width;        /* maxX - minX value */
    int height;       /* maxY - minY value */

    /* Device current state description */
    int x;
    int y;
    int z;
    int state;

} EloDeviceRec, *EloDevicePtr;

#endif /* __ELO_H__ */
