#include "elo_usb.h"

static void
EloUsbPtrControl(DeviceIntPtr pDev, PtrCtrl * ctrl)
{
    /* Nothing */
}

Bool
EloUsbControl(DeviceIntPtr pDev, int mode)
{
    InputInfoPtr pInfo;
    EloDevicePtr pPriv;
    Atom buttonLabel;
    Atom axisLabels[2] = { 0, 0 };
    unsigned char map[] = { 0, 1 };

    if (pDev == NULL)
        return !Success;

    pInfo = (InputInfoPtr) pDev->public.devicePrivate;
    pPriv = (EloDevicePtr) pInfo->private;

    switch (mode) {
    case DEVICE_INIT:

        xf86Msg(X_INFO, "%s: touchscreen init\n", pInfo->name);

        if (pPriv->screenNumber >= screenInfo.numScreens ||
            pPriv->screenNumber < 0)
            pPriv->screenNumber = 0;

        pPriv->screenWidth = screenInfo.screens[pPriv->screenNumber]->width;
        pPriv->screenHeight = screenInfo.screens[pPriv->screenNumber]->height;

         /**
         *
         * ---=== Init one button ===---
         *
         */

        if (InitButtonClassDeviceStruct(pDev, 1, &buttonLabel, map) == FALSE) {
            xf86Msg(X_ERROR, "%s: InitButtonClassDeviceStruct failed\n",
                    pInfo->name);
            return !Success;
        }

        if (pDev->focus == FALSE) {
            if (InitFocusClassDeviceStruct(pDev) == FALSE) {
                xf86Msg(X_ERROR, "%s: InitFocusClassDeviceStruct failed\n",
                        pInfo->name);
                return !Success;
            }
        }

        if (InitPtrFeedbackClassDeviceStruct(pDev, EloUsbPtrControl) == FALSE) {
            xf86Msg(X_ERROR, "%s: InitPtrFeedbackClassDeviceStruct failed\n",
                    pInfo->name);
            return !Success;
        }

        /**
         *
         * ---=== Init X and Y axes ===---
         *
         * Device reports motions on 2 axes in absolute coordinates.
         * Axes min and max values are reported in raw coordinates.
         * Resolution is computed roughly by the difference between
         * max and min values scaled from the approximate size
         * of the screen to fit one meter.
         */

        if (InitValuatorClassDeviceStruct(pDev,
                                          2,
                                          axisLabels,
                                          GetMotionHistorySize(),
                                          Absolute) == FALSE) {
            xf86Msg(X_ERROR, "%s: InitValuatorClassDeviceStruct failed\n",
                    pInfo->name);
            return !Success;
        }

        /* X */
        InitValuatorAxisStruct(pDev,
                               0,
                               axisLabels[0],
                               pPriv->minX,
                               pPriv->maxX,
                               9500,
                               0,       /* min resolution */
                               9500,    /* max resolution */
                               Absolute);

        /* Y */
        InitValuatorAxisStruct(pDev,
                               1,
                               axisLabels[1],
                               pPriv->minY,
                               pPriv->maxY,
                               10500,
                               0,      /* min resolution */
                               10500,   /* max resolution */
                               Absolute);

        if (pDev->focus == FALSE) {
            if (InitFocusClassDeviceStruct(pDev) == FALSE) {
                xf86Msg(X_ERROR, "%s: InitFocusClassDeviceStruct failed\n",
                        pInfo->name);
                return !Success;
            }
        }

        /* Allocate motion events buffer */
        xf86MotionHistoryAllocate(pInfo);

        break;

    case DEVICE_ON:

        xf86Msg(X_INFO, "%s: touchscreen on\n", pInfo->name);

        if (pDev->public.on == TRUE)
            break;

        if (pInfo->fd < 0) {
            /* Try to open device */
            SYSCALL(pInfo->fd = open(pPriv->device, O_RDWR | O_NONBLOCK));
            if (pInfo->fd < 0) {
                xf86Msg(X_ERROR, "%s: Unable to open touchscreen device %s\n",
                        pInfo->name, pPriv->device);
                return !Success;
            }

            unsigned short id[4] = { 0, 0, 0, 0 };
            char name[256] = "Unknown";

            /* Get device name */
            ioctl(pInfo->fd, EVIOCGNAME(sizeof(name)), name);
            xf86Msg(X_INFO, "%s: Device name: \"%s\"\n", pInfo->name, name);

            /* Get device id */
            ioctl(pInfo->fd, EVIOCGID, id);
            xf86Msg(X_INFO,
                    "%s: Device id: bus = 0x%x, vendor = 0x%x, \
                    product = 0x%x, version = 0x%x\n",
                    pInfo->name, id[ID_BUS], id[ID_VENDOR], id[ID_PRODUCT],
                    id[ID_VERSION]);
        }

        xf86FlushInput(pInfo->fd);
        xf86AddEnabledDevice(pInfo);
        pDev->public.on = TRUE;

        break;

        /*
         * Deactive the device. After this, the device will not emit events
         * until a subsequent DEVICE_ON. Device can be closed.
         */
    case DEVICE_OFF:

        xf86Msg(X_INFO, "%s: touchscreen off\n", pInfo->name);

        pDev->public.on = FALSE;

        if (pInfo->fd >= 0)
            xf86RemoveEnabledDevice(pInfo);

        SYSCALL(close(pInfo->fd));
        pInfo->fd = -1;
        break;

        /*
         * Final close before server exit. This is used during server shutdown.
         * Close device and free all resources.
         */
    case DEVICE_CLOSE:

        xf86Msg(X_INFO, "%s: touchscreen close\n", pInfo->name);

        pDev->public.on = FALSE;

        if (pInfo->fd >= 0)
            xf86RemoveEnabledDevice(pInfo);

        SYSCALL(close(pInfo->fd));
        pInfo->fd = -1;
        break;

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) * 100 + \
        GET_ABI_MINOR(ABI_XINPUT_VERSION) >= 1901
    case DEVICE_ABORT:
        xf86Msg(X_INFO, "%s: touchscreen abort\n", pInfo->name);
        break;
#endif

    default:
        xf86Msg(X_ERROR, "%s: Unsupported mode = %d\n", pInfo->name, mode);
        return BadValue;
        break;

    }

    return Success;
}

void
EloUsbReadInput(InputInfoPtr pInfo)
{
    EloDevicePtr pPriv;
    struct input_event events[ELO_NUM_EVENTS];
    int count;
    int ret;
    int x;
    int y;

    pPriv = (EloDevicePtr) pInfo->private;
    if (pPriv == NULL)
        return;

    while (xf86WaitForInput(pInfo->fd, 0) > 0) {
        SYSCALL(ret =
                read(pInfo->fd, &events,
                     sizeof(struct input_event) * ELO_NUM_EVENTS));

        if (ret <= 0) {
            /*
             * ENXIO  - No such device.
             * EAGAIN - Resource temporarily unavailable,
             * let`s try again later.
             */
            if (errno == ENXIO) {
                xf86Msg(X_ERROR,
                        "%s: Device disappeared! Disabling driver...\n",
                        pInfo->name);
                xf86RemoveEnabledDevice(pInfo);
                close(pInfo->fd);
                pInfo->fd = -1;
            }
            else if (errno != EAGAIN)
                xf86Msg(X_ERROR, "%s: Read error: %s\n", pInfo->name,
                        strerror(errno));

            return;
        }

        count = ret / sizeof(struct input_event);
        for (int i = 0; i < count; i++) {
            if (pPriv->debug == TRUE) {
                xf86Msg(X_INFO, "%s: type = %d, code = %d, value = %d\n",
                        pInfo->name, events[i].type, events[i].code,
                        events[i].value);
            }

            if (events[i].type == EV_KEY && events[i].code == BTN_TOUCH) {
                switch (events[i].value) {
                case TRUE:     /* button down */
                    pPriv->state = ELO_STATE_PRESS;
                    break;

                case FALSE:    /* button up */
                    pPriv->state = ELO_STATE_RELEASE;
                    break;
                }
            }
            else if (events[i].type == EV_ABS) {
                switch (events[i].code) {
                case ABS_X:    /* X absolute value */
                    pPriv->x = events[i].value;
                    break;

                case ABS_Y:    /* Y absolute value */
                    pPriv->y = events[i].value;
                    break;

                case ABS_PRESSURE:     /* Z absolute value */
                    pPriv->z = events[i].value;
                    break;
                }
            }
            /* Synchronization event is used to separate input events into packets */
            else if (events[i].type == EV_SYN) {
                if (pPriv->reverseX == TRUE)
                    x = pPriv->width - pPriv->x;
                else
                    x = pPriv->x;

                if (pPriv->reverseY == TRUE)
                    y = pPriv->height - pPriv->y;
                else
                    y = pPriv->y;

                xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);

                if ((pPriv->state == ELO_STATE_PRESS) &&
                    (pPriv->z > pPriv->threshold)) {
                    xf86PostButtonEvent(pInfo->dev, TRUE, 1, TRUE, 0, 2, x, y);
                }
                else if ((pPriv->state == ELO_STATE_RELEASE) ||
                         (pPriv->z <= pPriv->threshold)) {
                    xf86PostButtonEvent(pInfo->dev, TRUE, 1, FALSE, 0, 2, x, y);
                }

                if (pPriv->debug == TRUE) {
                    xf86Msg(X_INFO, "%s, x = %d, y = %d, z = %d, state = %s\n",
                            pInfo->name, pPriv->x, pPriv->y, pPriv->z,
                            pPriv->state == ELO_STATE_PRESS ? "down" : "up");
                }
            }
        }
    }
}
